@@ -42,7 +42,8 @@ enum {
 	SC_CARD_TYPE_CARDOS_M4_01,
 	SC_CARD_TYPE_CARDOS_M4_2,
 	SC_CARD_TYPE_CARDOS_M4_3,
-	SC_CARD_TYPE_CARDOS_M4_2B, /* 4.2b is the latest after 4.3b */
+	SC_CARD_TYPE_CARDOS_M4_2B, /* 4.2b is after 4.3b */
+	SC_CARD_TYPE_CARDOS_M4_2C,
 
 	/* flex/cyberflex drivers */
 	SC_CARD_TYPE_FLEX_BASE = 2000,
@@ -91,6 +92,8 @@ enum {
 	/* tcos driver */
 	SC_CARD_TYPE_TCOS_BASE = 8000,
 	SC_CARD_TYPE_TCOS_GENERIC,
+	SC_CARD_TYPE_TCOS_V2,
+	SC_CARD_TYPE_TCOS_V3,
 
 	/* openpgp driver */
 	SC_CARD_TYPE_OPENPGP_BASE = 9000,
@@ -135,7 +138,11 @@ enum {
 
 	/* TUBITAK UEKAE cards */
 	SC_CARD_TYPE_AKIS_BASE = 18000,
-	SC_CARD_TYPE_AKIS_GENERIC
+	SC_CARD_TYPE_AKIS_GENERIC,
+
+	/* EnterSafe cards */
+	SC_CARD_TYPE_ENTERSAFE_BASE = 19000,
+	SC_CARD_TYPE_ENTERSAFE_3K,
 };
 
 #ifdef __cplusplus
